from django import VERSION
from django.conf.urls import url, include

try:
    from django.conf.urls import patterns

except ImportError:
    patterns = None


urlpatterns = [
    url(r'^my_named_url/$', lambda r: None, name='named_url'),
]


def get_mock_patterns():
    url_ = url(r'^my_another_named_url/$', lambda r: None, name='url')

    if patterns:
        return patterns('', url_)

    return [url_]


class MockUrlconfModule(object):

    urlpatterns = get_mock_patterns()


if VERSION < (2, 0):
    urlpatterns.append(url(r'^namespace/', include((MockUrlconfModule, None, 'namespaced'))))
else:
    urlpatterns.append(url(r'^namespace/', include((MockUrlconfModule, 'app'), namespace='namespaced')))


if VERSION < (1, 10):
    urlpatterns.insert(0, '')
    urlpatterns = patterns(*urlpatterns)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Block',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('alias', models.CharField(help_text='Short name to address this block from a template.', max_length=80, verbose_name='Alias', db_index=True)),
                ('description', models.CharField(help_text='Short memo for this block.', max_length=100, null=True, verbose_name='Description', blank=True)),
                ('url', models.CharField(help_text='Page URL this block is related to. Regular expressions supported (e.g.: "/news.*" &#8212; everything under "/news").<br />View names are supported: prepend <b>:</b> to view name <br /><b>Reserved URL alias:</b> "*" &#8212; every page.', max_length=200, verbose_name='URL')),
                ('contents', models.TextField(help_text='Block contents to render in a template.', verbose_name='Contents')),
                ('hidden', models.BooleanField(default=False, help_text='Whether to show this block when requested.', db_index=True, verbose_name='Hidden')),
            ],
            options={
                'verbose_name': 'Site Block',
                'verbose_name_plural': 'Site Blocks',
            },
            bases=(models.Model,),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Block(models.Model):

    alias = models.CharField(
        _('Alias'), max_length=80, help_text=_('Short name to address this block from a template.'), db_index=True)
    description = models.CharField(
        _('Description'), max_length=100, help_text=_('Short memo for this block.'), null=True, blank=True)
    url = models.CharField(
        _('URL'), max_length=200,
        help_text=_('Page URL this block is related to. Regular expressions supported (e.g.: "/news.*" &#8212; '
                    'everything under "/news").<br />'
                    'View names are supported: prepend <b>:</b> to view name <br />'
                    '<b>Reserved URL alias:</b> "*" &#8212; every page.'))
    contents = models.TextField(
        _('Contents'), help_text=_('Block contents to render in a template.'))
    access_loggedin = models.BooleanField(
        _('Logged in only'),
        help_text=_('Check it to grant access to this block to authenticated users only.'),
        db_index=True, default=False)
    access_guest = models.BooleanField(
        _('Guests only'),
        help_text=_('Check it to grant access to this block to guests only.'), db_index=True, default=False)
    hidden = models.BooleanField(
        _('Hidden'), help_text=_('Whether to show this block when requested.'), db_index=True, default=False)
      
    class Meta(object):
        verbose_name = _('Site Block')
        verbose_name_plural = _('Site Blocks')
        
    def __str__(self):
        return self.alias



import re
from collections import defaultdict
from random import choice

from django import VERSION
from django.core.cache import cache

try:
    from django.urls import resolve, Resolver404
except ImportError:
    from django.core.urlresolvers import resolve, Resolver404

from django.utils.translation import get_language
from django.db.models import signals

from .models import Block
from .settings import I18N_SUPPORT


if I18N_SUPPORT:
    cache_get_key = lambda block_alias: '%s:%s' % (block_alias, get_language())
else:
    cache_get_key = lambda block_alias: block_alias


DJANGO_2 = VERSION >= (2, 0, 0)


# Contains dynamic blocks.
_DYNAMIC_BLOCKS = defaultdict(list)


def register_dynamic_block(alias, callable):
    """Registers a callable that produces contents for a dynamic block.

    Callable on call will get the following kwargs:

        * `block_alias` - block alias,
        * `block_context` - template context for block

    Example::

        # Put the following code somewhere where it'd be triggered as expected. E.g. in app view.py.

        from random import choice
        # Import the register function.
        from siteblocks.siteblocksapp import register_dynamic_block


        # The following function will be used as a block contents producer.
        def get_quote(**kwargs):
            quotes = [  # From Terry Pratchett's Discworld novels.
                'Ripples of paradox spread out across the sea of causality.',
                'Early to rise, early to bed, makes a man healthy, wealthy and dead.',
                'Granny had nothing against fortune-telling provided it was done badly by people with no talent for it.'
                'Take it from me, there\'s nothing more terrible than someone out to do the world a favour.',
                'The duke had a mind that ticked like a clock and, like a clock, it regularly went cuckoo.',
                'Most gods find it hard to walk and think at the same time.',
                'They didn\'t have to be funny - they were father jokes',
                'Speak softly and employ a huge man with a crowbar.',
            ]
            return choice(quotes)

        # And we register our siteblock.
        register_dynamic_block('quote', get_quote)

    """
    global _DYNAMIC_BLOCKS
    _DYNAMIC_BLOCKS[alias].append(callable)


def get_dynamic_blocks():
    """Returns a dictionary with currently registered dynamic blocks."""
    return _DYNAMIC_BLOCKS


class SiteBlocks(object):

    # siteblocks objects are stored in Django cache for a year (60 * 60 * 24 * 365 = 31536000 sec).
    # Cache is only invalidated on block item change.
    CACHE_TIMEOUT = 31536000
    CACHE_KEY = 'siteblocks'

    IDX_GUEST = 0
    IDX_AUTH = 1

    def __init__(self):
        self._cache = None
        signals.post_save.connect(self._cache_empty, sender=Block)
        signals.post_delete.connect(self._cache_empty, sender=Block)

    def _cache_init(self):
        """Initializes local cache from Django cache."""
        cache_ = cache.get(self.CACHE_KEY)
        if cache_ is None:
            cache_ = defaultdict(dict)
        self._cache = cache_

    def _cache_save(self):
        cache.set(self.CACHE_KEY, self._cache, self.CACHE_TIMEOUT)

    def _cache_get(self, key):
        """Returns cache entry parameter value by its name."""
        return self._cache.get(key, False)

    def _cache_set(self, key, value):
        """Replaces entire cache entry parameter data by its name with new data."""
        self._cache[key] = value

    def _cache_empty(self, **kwargs):
        self._cache = None
        cache.delete(self.CACHE_KEY)

    def get_contents_static(self, block_alias, context):
        """Returns contents of a static block."""

        if 'request' not in context:
            # No use in further actions as we won't ever know current URL.
            return ''

        current_url = context['request'].path

        # Resolve current view name to support view names as block URLs.
        try:
            resolver_match = resolve(current_url)
            namespace = ''
            if resolver_match.namespaces:
                # More than one namespace, really? Hmm.
                namespace = resolver_match.namespaces[0]
            resolved_view_name = ':%s:%s' % (namespace, resolver_match.url_name)
        except Resolver404:
            resolved_view_name = None

        self._cache_init()

        cache_entry_name = cache_get_key(block_alias)

        siteblocks_static = self._cache_get(cache_entry_name)
        if not siteblocks_static:
            blocks = Block.objects.filter(alias=block_alias, hidden=False).only('url', 'contents')
            siteblocks_static = [defaultdict(list), defaultdict(list)]
            for block in blocks:
                if block.url == '*':
                    url_re = block.url
                elif block.url.startswith(':'):
                    url_re = block.url
                    # Normalize URL name to include namespace.
                    if url_re.count(':') == 1:
                        url_re = ':%s' % url_re
                else:
                    url_re = re.compile(r'%s' % block.url)

                if block.access_guest:
                    siteblocks_static[self.IDX_GUEST][url_re].append(block.contents)
                elif block.access_loggedin:
                    siteblocks_static[self.IDX_AUTH][url_re].append(block.contents)
                else:
                    siteblocks_static[self.IDX_GUEST][url_re].append(block.contents)
                    siteblocks_static[self.IDX_AUTH][url_re].append(block.contents)

            self._cache_set(cache_entry_name, siteblocks_static)

        self._cache_save()

        user = getattr(context['request'], 'user', None)

        is_authenticated = getattr(user, 'is_authenticated', False)

        if not DJANGO_2:
            is_authenticated = is_authenticated()

        if is_authenticated:
            lookup_area = siteblocks_static[self.IDX_AUTH]
        else:
            lookup_area = siteblocks_static[self.IDX_GUEST]

        static_block_contents = ''
        if '*' in lookup_area:
            static_block_contents = choice(lookup_area['*'])
        elif resolved_view_name in lookup_area:
            static_block_contents = choice(lookup_area[resolved_view_name])
        else:
            for url, contents in lookup_area.items():
                if url.match(current_url):
                    static_block_contents = choice(contents)
                    break

        return static_block_contents

    def get_contents_dynamic(self, block_alias, context):
        """Returns contents of a dynamic block."""
        dynamic_block = get_dynamic_blocks().get(block_alias, [])
        if not dynamic_block:
            return ''

        dynamic_block = choice(dynamic_block)
        return dynamic_block(block_alias=block_alias, block_context=context)

    def get(self, block_alias, context):
        """Main method returning block contents (static or dynamic)."""
        contents = []

        dynamic_block_contents = self.get_contents_dynamic(block_alias, context)
        if dynamic_block_contents:
            contents.append(dynamic_block_contents)

        static_block_contents = self.get_contents_static(block_alias, context)
        if static_block_contents:
            contents.append(static_block_contents)

        if not contents:
            return ''

        return choice(contents)



from django import template
from django.template.base import FilterExpression

from ..siteblocksapp import SiteBlocks

register = template.Library()

# Utility methods are implemented in SiteBlocks class
siteblocks = SiteBlocks()


@register.tag
def siteblock(parser, token):
    """Two notation types are acceptable:

        1. Two arguments:
           {% siteblock "myblock" %}
           Used to render "myblock" site block.
           
        2. Four arguments:
           {% siteblock "myblock" as myvar %}
           Used to put "myblock" site block into "myvar" template variable.
           
    """
    tokens = token.split_contents()
    tokens_num = len(tokens)

    if tokens_num not in (2, 4):
        raise template.TemplateSyntaxError(
            '%r tag requires two or four arguments. '
            'E.g.: {%% siteblock "myblock" %%} or {%% siteblock "myblock" as myvar %%}.' % tokens[0])

    block_alias = parser.compile_filter(tokens[1])
    as_var = None
    tokens = tokens[2:]
    if len(tokens) >= 2 and tokens[-2] == 'as':
        as_var = tokens[-1]

    return siteblockNode(block_alias, as_var)


class siteblockNode(template.Node):

    def __init__(self, block_alias, as_var=None):
        self.block_alias = block_alias
        self.as_var = as_var
        
    def render(self, context):
        block_alias = self.block_alias
        if isinstance(self.block_alias, FilterExpression):
            block_alias = block_alias.resolve(context)

        contents = siteblocks.get(block_alias, context)

        if self.as_var is not None:
            context[self.as_var] = contents
            return ''

        return contents



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SiteblocksConfig(AppConfig):
    """Siteblocks configuration."""

    name = 'siteblocks'
    verbose_name = _('Site Blocks')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('siteblocks', '0002_auto_20140921_1038'),
    ]

    operations = [
        migrations.AlterField(
            model_name='block',
            name='access_guest',
            field=models.BooleanField(default=False, help_text='Check it to grant access to this block to guests only.', db_index=True, verbose_name='Guests only'),
        ),
        migrations.AlterField(
            model_name='block',
            name='access_loggedin',
            field=models.BooleanField(default=False, help_text='Check it to grant access to this block to authenticated users only.', db_index=True, verbose_name='Logged in only'),
        ),
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import Block


class BlockAdmin(admin.ModelAdmin):

    list_display = ('alias', 'description', 'url', 'hidden', 'access_loggedin', 'access_guest')
    search_fields = ['alias', 'url']
    list_filter = ['hidden']
    ordering = ['alias']
    fieldsets = (
        (_('Basic settings'), {
            'fields': ('alias', 'url', 'contents',)
        }),
        (_('Access settings'), {
            'classes': ('collapse',),
            'fields': ('access_loggedin', 'access_guest')
        }),
        (_('Additional settings'), {
            'fields': ('description', 'hidden')
        }),
    )


admin.site.register(Block, BlockAdmin)



from random import choice

from siteblocks.models import Block
from siteblocks.siteblocksapp import SiteBlocks, register_dynamic_block


QUOTES = [  # From Terry Pratchett's Discworld novels.
    'Ripples of paradox spread out across the sea of causality.',
    'Early to rise, early to bed, makes a man healthy, wealthy and dead.',
    'Granny had nothing against fortune-telling provided it was done badly by people with no talent for it.',
    'Take it from me, there\'s nothing more terrible than someone out to do the world a favour.',
    'The duke had a mind that ticked like a clock and, like a clock, it regularly went cuckoo.',
    'Most gods find it hard to walk and think at the same time.',
    'They didn\'t have to be funny - they were father jokes',
    'Speak softly and employ a huge man with a crowbar.',
]


def get_quote(**kwargs):
    return choice(QUOTES)


class TestTemplateTags(object):

    def test_siteblock(self, template_render_tag, template_context):

        context = template_context({})

        assert template_render_tag('siteblocks', 'siteblock "myblock"', context) == ''

        b1 = Block(alias='myblock', url='*', contents='my_block_here')
        b1.save(force_insert=True)

        assert template_render_tag('siteblocks', 'siteblock "myblock"', context) == 'my_block_here'

        template_render_tag('siteblocks', 'siteblock "myblock" as somevar', context)
        assert context.get('somevar') == 'my_block_here'


class TestTreeItemModel(object):

    @classmethod
    def setup_method(cls):
        cls.siteblocks = SiteBlocks()

        cls.b1 = Block(alias='main', url='*', contents='main_every_visible')
        cls.b1.save(force_insert=True)

        cls.b2 = Block(alias='main', description='hidden', url='*', contents='main_every_hidden', hidden=True)
        cls.b2.save(force_insert=True)

        cls.b3 = Block(alias='multiple', url='*', contents='multiple_1')
        cls.b3.save(force_insert=True)

        cls.b4 = Block(alias='multiple', url='*', contents='multiple_2')
        cls.b4.save(force_insert=True)

        cls.b5 = Block(alias='filtered_1', url='/news.*', contents='filtered_1_1')
        cls.b5.save(force_insert=True)

        cls.b6 = Block(alias='filtered_1', url='/gro{1,2}ves', contents='filtered_1_2')
        cls.b6.save(force_insert=True)

        cls.b7 = Block(alias='named_1', url=':named_url', contents='named_1_1')
        cls.b7.save(force_insert=True)

        cls.b8 = Block(alias='named_2', url=':namespaced:url', contents='named_2_1')
        cls.b8.save(force_insert=True)

        cls.b9 = Block(alias='access_filter', url='*', contents='afilter_auth', access_loggedin=True)
        cls.b9.save(force_insert=True)

        cls.b10 = Block(alias='access_filter', url='*', contents='afilter_guest', access_guest=True)
        cls.b10.save(force_insert=True)

    def test_static_notalias(self, template_context):

        contents = self.siteblocks.get('notalias', template_context({}, '/root/'))
        assert contents == ''

    def test_static_asterisk(self, template_context):

        contents = self.siteblocks.get(self.b1.alias, template_context({}, '/root/'))
        assert contents == self.b1.contents

        contents = self.siteblocks.get(self.b3.alias, template_context({}, '/root/'))
        assert contents in [self.b3.contents, self.b4.contents]

    def test_static_multiple(self, template_context):

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/news/1'))
        assert contents == self.b5.contents

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/news/2'))
        assert contents == self.b5.contents

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/new/2'))
        assert contents == ''

    def test_static_regexp(self, template_context):

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/grooves'))
        assert contents == self.b6.contents

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/groves'))
        assert contents == self.b6.contents

        contents = self.siteblocks.get(self.b5.alias, template_context({}, '/groooves'))
        assert contents == ''

    def test_static_namedview(self, template_context):

        contents = self.siteblocks.get(self.b7.alias, template_context({}, '/my_named_url/'))
        assert contents == self.b7.contents

        contents = self.siteblocks.get(self.b8.alias, template_context({}, '/namespace/my_another_named_url/'))
        assert contents == self.b8.contents

    def test_dynamic(self, template_context):
        register_dynamic_block('quotes', get_quote)
        contents = self.siteblocks.get('quotes', template_context({}, '/somewhere/'))
        assert contents in QUOTES

    def test_static_access_filters(self, template_context, user_create):
        contents = self.siteblocks.get(self.b9.alias, template_context({}, '/some/'))
        assert contents == self.b10.contents

        contents = self.siteblocks.get(self.b9.alias, template_context({}, '/some/', user=user_create()))
        assert contents == self.b9.contents
